<?php
/**
 * Plugin 'rdt_modalbox2' for the 'ameos_formidable' extension.
 *
 * @author  Jerome Schneider <typo3dev@ameos.com>
 */
class tx_mkforms_widgets_modalbox2_Main extends formidable_mainrenderlet
{
    public $aLibs = [
        'rdt_modalbox2_class' => 'Resources/Public/JavaScript/widgets/modalbox2/modalbox2.js',
        'rdt_modalbox2_lib_class' => 'Resources/Public/JavaScript/widgets/modalbox2/modalbox1.6.0/modalbox.js',
    ];

    public $bCustomIncludeScript = true;
    public $sMajixClass = 'ModalBox2';

    public function _render()
    {
        // allowed because of $bCustomIncludeScript = TRUE
        $this->includeScripts();

        return '';
    }

    public function _renderReadOnly()
    {
        return $this->_render();
    }

    public function _readOnly()
    {
        return true;
    }

    public function _renderOnly($bForAjax = false)
    {
        return true;
    }

    public function mayHaveChilds()
    {
        return true;
    }

    public function majixShowBox($aConfig = [], $aTags = [])
    {
        if ('EID' !== tx_mkforms_util_Div::getEnvExecMode()) {
            $aEventsBefore = array_keys($this->oForm->aRdtEvents);
        }

        $aChildsBag = $this->renderChildsBag();
        $aChildsBag = \Sys25\RnBase\Utility\Arrays::mergeRecursiveWithOverrule($aChildsBag, $aTags);

        if ('EID' !== tx_mkforms_util_Div::getEnvExecMode()) {
            $aEventsAfter = array_keys($this->oForm->aRdtEvents);
            $aAddedKeys = array_diff($aEventsAfter, $aEventsBefore);
            $aAddedEvents = [];
            foreach ($aAddedKeys as $sKey) {
                $aAddedEvents[$sKey] = $this->oForm->aRdtEvents[$sKey];
                unset($this->oForm->aRdtEvents[$sKey]);
                // unset because if rendered in a lister,
                // we need to be able to detect the new events even if they were already declared by other loops in the lister
            }

            $aConfig['attachevents'] = $aAddedEvents;
            $aConfig['postinit'] = $this->oForm->aPostInitTasks;
        } else {
            // specific to this renderlet
            // as events have to be attached to the HTML
            // after the execution of the majix tasks
            // in that case, using the modalbox's afterLoad event handler
            $aConfig['attachevents'] = $this->oForm->aRdtEventsAjax;
            $aConfig['postinit'] = $this->oForm->aPostInitTasksAjax;
            $this->oForm->aRdtEventsAjax = [];
            $this->oForm->aPostInitTasksAjax = [];
        }

        $sCompiledChilds = $this->renderChildsCompiled(
            $aChildsBag
        );

        $aConfig['html'] = $sCompiledChilds;

        return $this->buildMajixExecuter(
            'showBox',
            $aConfig
        );
    }

    public function majixCloseBox($aOptions = false)
    {
        return $this->buildMajixExecuter(
            'closeBox',
            $aOptions
        );
    }

    public function majixRepaint()
    {
        return $this->buildMajixExecuter(
            'repaint',
            $this->renderChildsCompiled(
                $this->renderChildsBag()
            )
        );
    }

    public function majixResizeToContent()
    {
        return $this->buildMajixExecuter(
            'resizeToContent'
        );
    }

    public function majixResizeToInclude($sHtmlId)
    {
        return $this->buildMajixExecuter(
            'resizeToInclude',
            $sHtmlId
        );
    }

    // this has to be static !!!
    public static function loaded(&$aParams)
    {
        $aParams['form']->getJSLoader()->loadScriptaculous();
        $sCss = $aParams['form']->toWebPath('EXT:mkforms/Resources/Public/JavaScript/widgets/modalbox2/modalbox1.6.0/modalbox.css');
        $aParams['form']->additionalHeaderData(
            '<link rel="stylesheet" type="text/css" href="'.$sCss.'" />'
        );
    }
}
